#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic Curve Schnorr Signature Algorithm (ECSSA).

This implementation is according to BIP340-Schnorr:

https://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki

Differently from ECDSA, the BIP340-Schnorr scheme supports
messages of size hsize only.

It also uses as public key the x-coordinate (field element)
of the curve point associated to the private key 0 < q < n.
Therefore, for sepcp256k1 the public key size is 32 bytes.
Arguably, the knowledge of q as the discrete logarithm of Q
also implies the knowledge of n-q as discrete logarithm of -Q.
As such, {q, n-q} can be considered a single private key and
{Q, -Q} the associated public key characterized by the shared x_Q.

Also, BIP340 advocates its own SHA256 modification as hash function:
TaggedHash(tag, x) = SHA256(SHA256(tag)||SHA256(tag)||x)
The rationale is to make BIP340 signatures invalid for anything else
but Bitcoin and vice versa.

TaggedHash is used for both the challenge (with tag 'BIPSchnorr')
and the deterministic nonce (with tag 'BIPSchnorrDerive').

To allow for secure batch verification of multiple signatures,
BIP340-Schnorr uses a challenge that prevents public key recovery
from signature: c = TaggedHash('BIPSchnorr', x_k||x_Q||msg).

A custom deterministic algorithm for the ephemeral key (nonce)
is used for signing, instead of the RFC6979 standard:
k = TaggedHash('BIPSchnorrDerive', q||msg)

Finally, BIP340-Schnorr adopts a robust [r][s] custom serialization
format, instead of the loosely specified ASN.1 DER standard.
The signature size is p-size*n-size, where p-size is the field element
(curve point coordinate) byte size and n-size is the scalar
(curve point multiplication coefficient) byte size.
For sepcp256k1 the resulting signature size is 64 bytes.
"""

import secrets
from dataclasses import InitVar, dataclass, field
from hashlib import sha256
from typing import List, Optional, Sequence, Tuple, Type, TypeVar, Union

from dataclasses_json import DataClassJsonMixin, config

from .alias import BinaryData, HashF, Integer, JacPoint, Octets, Point, String
from .bip32 import BIP32Key
from .curve import Curve, secp256k1
from .curve_group import _double_mult, _mult, _multi_mult
from .exceptions import BTClibRuntimeError, BTClibTypeError, BTClibValueError
from .hashes import reduce_to_hlen, tagged_hash
from .number_theory import mod_inv
from .to_prv_key import PrvKey, int_from_prv_key
from .to_pub_key import point_from_pub_key
from .utils import (
    bytes_from_octets,
    bytesio_from_binarydata,
    hex_string,
    int_from_bits,
)

_Sig = TypeVar("_Sig", bound="Sig")


@dataclass(frozen=True)
class Sig(DataClassJsonMixin):
    """BIP340-Schnorr signature.

    r is a _field_element_, 0 <= r < ec.p
    s is a scalar, 0 <= s < ec.n (yes, for BIP340-Schnorr it can be zero)
    (p is the field prime, n is the curve order)
    """

    # 32 bytes
    r: int = field(
        default=-1, metadata=config(encoder=lambda v: v.hex(), decoder=bytes.fromhex)
    )
    # 32 bytes
    s: int = field(
        default=-1, metadata=config(encoder=lambda v: v.hex(), decoder=bytes.fromhex)
    )
    ec: Curve = field(
        default=secp256k1,
        metadata=config(encoder=lambda v: v.name(), decoder=bytes.fromhex),
    )
    check_validity: InitVar[bool] = True

    def __post_init__(self, check_validity: bool) -> None:
        if check_validity:
            self.assert_valid()

    def assert_valid(self) -> None:
        # Fail if r is not a field element, i.e. not a valid x-coordinate
        self.ec.y(self.r)

        # Fail if s is not [0, n-1].
        if not 0 <= self.s < self.ec.n:
            err_msg = "scalar s not in 0..n-1: "
            err_msg += f"'{hex_string(self.s)}'" if self.s > 0xFFFFFFFF else f"{self.s}"
            raise BTClibValueError(err_msg)

    def serialize(self, check_validity: bool = True) -> bytes:

        if check_validity:
            self.assert_valid()

        out = self.r.to_bytes(self.ec.psize, byteorder="big", signed=False)
        out += self.s.to_bytes(self.ec.nsize, byteorder="big", signed=False)
        return out

    @classmethod
    def deserialize(
        cls: Type[_Sig], data: BinaryData, check_validity: bool = True
    ) -> _Sig:

        stream = bytesio_from_binarydata(data)
        ec = secp256k1
        r = int.from_bytes(stream.read(ec.psize), byteorder="big", signed=False)
        s = int.from_bytes(stream.read(ec.nsize), byteorder="big", signed=False)
        return cls(r, s, ec, check_validity)


# hex-string or bytes representation of an int
# 33 or 65 bytes or hex-string
# BIP32Key as dict or String
# tuple Point
BIP340PubKey = Union[Integer, Octets, BIP32Key, Point]


def point_from_bip340pub_key(x_Q: BIP340PubKey, ec: Curve = secp256k1) -> Point:
    """Return a verified-as-valid BIP340 public key as Point tuple.

    It supports:

    - BIP32 extended keys (bytes, string, or BIP32KeyData)
    - SEC Octets (bytes or hex-string, with 02, 03, or 04 prefix)
    - BIP340 Octets (bytes or hex-string, p-size Point x-coordinate)
    - native tuple
    """

    # BIP 340 key as integer
    if isinstance(x_Q, int):
        return x_Q, ec.y_even(x_Q)

    # (tuple) Point, (dict or str) BIP32Key, or 33/65 bytes
    try:
        x_Q = point_from_pub_key(x_Q, ec)[0]
        return x_Q, ec.y_even(x_Q)
    except BTClibValueError:
        pass

    # BIP 340 key as bytes or hex-string
    if isinstance(x_Q, (str, bytes)):
        Q = bytes_from_octets(x_Q, ec.psize)
        x_Q = int.from_bytes(Q, "big", signed=False)
        return x_Q, ec.y_even(x_Q)

    raise BTClibTypeError("not a BIP340 public key")


def gen_keys_(
    prv_key: PrvKey = None, ec: Curve = secp256k1
) -> Tuple[int, int, JacPoint]:
    "Return a BIP340 private/public (int, JacPoint) key-pair."

    if prv_key is None:
        q = 1 + secrets.randbelow(ec.n - 1)
    else:
        q = int_from_prv_key(prv_key, ec)

    QJ = _mult(q, ec.GJ, ec)
    x_Q, y_Q = ec._aff_from_jac(QJ)
    if y_Q % 2:
        q = ec.n - q
        QJ = ec.negate_jac(QJ)

    return q, x_Q, QJ


def gen_keys(prv_key: PrvKey = None, ec: Curve = secp256k1) -> Tuple[int, int]:
    "Return a BIP340 private/public (int, int) key-pair."

    if prv_key is None:
        q = 1 + secrets.randbelow(ec.n - 1)
    else:
        q = int_from_prv_key(prv_key, ec)

    QJ = _mult(q, ec.GJ, ec)
    x_Q, y_Q = ec._aff_from_jac(QJ)
    if y_Q % 2:
        q = ec.n - q

    return q, x_Q


def __det_nonce(m: bytes, q: int, Q: int, aux: bytes, ec: Curve, hf: HashF) -> int:

    # assume the random oracle model for the hash function,
    # i.e. hash values can be considered uniformly random

    # Note that in general, taking a uniformly random integer
    # modulo the curve order n would produce a biased result.
    # However, if the order n is sufficiently close to 2^hlen,
    # then the bias is not observable:
    # e.g. for secp256k1 and sha256 1-n/2^256 it is about 1.27*2^-128
    #
    # the unbiased implementation is provided here,
    # which works also for very-low-cardinality test curves

    randomizer = tagged_hash("BIP0340/aux", aux, hf)
    xor = q ^ int.from_bytes(randomizer, "big", signed=False)
    max_len = max(ec.nsize, hf().digest_size)
    t = xor.to_bytes(max_len, byteorder="big", signed=False)

    t += Q.to_bytes(ec.psize, byteorder="big", signed=False) + m

    while True:
        t = tagged_hash("BIP0340/nonce", t, hf)
        # The following lines would introduce a bias
        # k = int.from_bytes(t, 'big') % ec.n
        # k = int_from_bits(t, ec.nlen) % ec.n
        k = int_from_bits(t, ec.nlen)  # candidate k
        if 0 < k < ec.n:  # acceptable value for k
            return k  # successful candidate


def _det_nonce(
    m: Octets,
    prv_key: PrvKey,
    aux: Optional[Octets] = None,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> int:
    """Return a BIP340 deterministic ephemeral key (nonce)."""

    # the message m: a hlen array
    hlen = hf().digest_size
    m = bytes_from_octets(m, hlen)

    q, Q = gen_keys(prv_key, ec)

    # the auxiliary random component
    aux = secrets.token_bytes(hlen) if aux is None else bytes_from_octets(aux)

    return __det_nonce(m, q, Q, aux, ec, hf)


def det_nonce(
    msg: String, prv_key: PrvKey, ec: Curve = secp256k1, hf: HashF = sha256
) -> int:
    """Return a BIP340 deterministic ephemeral key (nonce)."""

    m = reduce_to_hlen(msg, hf)
    return _det_nonce(m, prv_key, None, ec, hf)


def __challenge(m: bytes, x_Q: int, x_K: int, ec: Curve, hf: HashF) -> int:

    t = x_K.to_bytes(ec.psize, byteorder="big", signed=False)
    t += x_Q.to_bytes(ec.psize, byteorder="big", signed=False)
    t += m
    t = tagged_hash("BIP0340/challenge", t, hf)
    c = int_from_bits(t, ec.nlen) % ec.n
    if c == 0:
        raise BTClibRuntimeError("invalid zero challenge")  # pragma: no cover
    return c


def _challenge(
    m: Octets,
    Q: BIP340PubKey,
    K: BIP340PubKey,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> int:

    # the message m: a hlen array
    hlen = hf().digest_size
    m = bytes_from_octets(m, hlen)

    x_Q, _ = point_from_bip340pub_key(Q, ec)
    x_K, _ = point_from_bip340pub_key(K, ec)

    return __challenge(m, x_Q, x_K, ec, hf)


def challenge(
    msg: String,
    Q: BIP340PubKey,
    K: BIP340PubKey,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> int:

    m = reduce_to_hlen(msg, hf)
    return _challenge(m, Q, K, ec, hf)


def __sign(c: int, q: int, k: int, r: int, ec: Curve) -> Sig:
    # Private function for testing purposes: it allows to explore all
    # possible value of the challenge c (for low-cardinality curves).
    # It assume that c is in [1, n-1], while q and k are in [1, n-1]

    if c == 0:  # c≠0 required as it multiplies the private key
        raise BTClibRuntimeError("invalid zero challenge")

    # s=0 is ok: in verification there is no inverse of s
    s = (k + c * q) % ec.n

    return Sig(r, s, ec)


def _sign(
    m: Octets,
    prv_key: PrvKey,
    k: Optional[PrvKey] = None,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> Sig:
    """Sign a hlen bytes message according to BIP340 signature algorithm.

    If the deterministic nonce is not provided,
    the BIP340 specification (not RFC6979) is used.
    """

    # the message m: a hlen array
    hlen = hf().digest_size
    m = bytes_from_octets(m, hlen)

    # private and public keys
    q, x_Q = gen_keys(prv_key, ec)

    # the nonce k: an integer in the range 1..n-1.
    if k is None:
        k = __det_nonce(m, q, x_Q, secrets.token_bytes(hlen), ec, hf)

    k, x_K = gen_keys(k, ec)

    # the challenge
    c = __challenge(m, x_Q, x_K, ec, hf)

    return __sign(c, q, k, x_K, ec)


def sign(
    msg: String, prv_key: PrvKey, ec: Curve = secp256k1, hf: HashF = sha256
) -> Sig:
    """Sign message according to BIP340 signature algorithm.

    The message msg is first processed by hf, yielding the value

        m = hf(msg),

    a sequence of bits of length *hlen*.

    Normally, hf is chosen such that its output length *hlen* is
    roughly equal to *nlen*, the bit-length of the group order *n*,
    since the overall security of the signature scheme will depend on
    the smallest of *hlen* and *nlen*; however, ECSSA
    supports all combinations of *hlen* and *nlen*.

    The BIP340 deterministic nonce (not RFC6979) is used.
    """

    m = reduce_to_hlen(msg, hf)
    return _sign(m, prv_key, None, ec, hf)


def __assert_as_valid(c: int, QJ: JacPoint, r: int, s: int, ec: Curve) -> None:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    # Let K = sG - eQ.
    # in Jacobian coordinates
    KJ = _double_mult(ec.n - c, QJ, s, ec.GJ, ec)

    # Fail if infinite(KJ).
    # Fail if y_K is odd.
    if ec._y_aff_from_jac(KJ) % 2:
        raise BTClibRuntimeError("y_K is odd")

    # Fail if x_K ≠ r
    if KJ[0] != KJ[2] * KJ[2] * r % ec.p:
        raise BTClibRuntimeError("signature verification failed")


def _assert_as_valid(
    m: Octets, Q: BIP340PubKey, sig: Union[Sig, Octets], hf: HashF = sha256
) -> None:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    if not isinstance(sig, Sig):
        sig = Sig.deserialize(sig)
    else:
        sig.assert_valid()  # 1

    x_Q, y_Q = point_from_bip340pub_key(Q, sig.ec)

    # Let c = int(hf(bytes(r) || bytes(Q) || m)) mod n.
    c = _challenge(m, x_Q, sig.r, sig.ec, hf)

    __assert_as_valid(c, (x_Q, y_Q, 1), sig.r, sig.s, sig.ec)


def assert_as_valid(
    msg: String, Q: BIP340PubKey, sig: Union[Sig, Octets], hf: HashF = sha256
) -> None:

    m = reduce_to_hlen(msg, hf)
    _assert_as_valid(m, Q, sig, hf)


def _verify(
    m: Octets, Q: BIP340PubKey, sig: Union[Sig, Octets], hf: HashF = sha256
) -> bool:
    """Verify the BIP340 signature of the provided message."""

    # all kind of Exceptions are catched because
    # verify must always return a bool
    try:
        _assert_as_valid(m, Q, sig, hf)
    except Exception:  # pylint: disable=broad-except
        return False
    else:
        return True


def verify(
    msg: String, Q: BIP340PubKey, sig: Union[Sig, Octets], hf: HashF = sha256
) -> bool:
    """ECDSA signature verification (SEC 1 v.2 section 4.1.4)."""

    m = reduce_to_hlen(msg, hf)
    return _verify(m, Q, sig, hf)


def __recover_pub_key(c: int, r: int, s: int, ec: Curve) -> int:
    # Private function provided for testing purposes only.

    if c == 0:
        raise BTClibRuntimeError("invalid zero challenge")

    KJ = r, ec.y_even(r), 1

    e1 = mod_inv(c, ec.n)
    QJ = _double_mult(ec.n - e1, KJ, e1 * s, ec.GJ, ec)
    # edge case that cannot be reproduced in the test suite
    if QJ[2] == 0:
        err_msg = "invalid (INF) key"  # pragma: no cover
        raise BTClibRuntimeError(err_msg)  # pragma: no cover
    return ec._x_aff_from_jac(QJ)


def _crack_prv_key(
    m_1: Octets,
    sig1: Union[Sig, Octets],
    m_2: Octets,
    sig2: Union[Sig, Octets],
    Q: BIP340PubKey,
    hf: HashF = sha256,
) -> Tuple[int, int]:

    if not isinstance(sig1, Sig):
        sig1 = Sig.deserialize(sig1)
    else:
        sig1.assert_valid()  # 1

    if not isinstance(sig2, Sig):
        sig2 = Sig.deserialize(sig2)
    else:
        sig2.assert_valid()  # 1

    ec = sig2.ec
    if sig1.ec != ec:
        raise BTClibValueError("not the same curve in signatures")
    if sig1.r != sig2.r:
        raise BTClibValueError("not the same r in signatures")
    if sig1.s == sig2.s:
        raise BTClibValueError("identical signatures")

    hlen = hf().digest_size
    m_1 = bytes_from_octets(m_1, hlen)
    m_2 = bytes_from_octets(m_2, hlen)

    x_Q = point_from_bip340pub_key(Q, ec)[0]

    c_1 = _challenge(m_1, x_Q, sig1.r, ec, hf)
    c_2 = _challenge(m_2, x_Q, sig2.r, ec, hf)
    q = (sig1.s - sig2.s) * mod_inv(c_2 - c_1, ec.n) % ec.n
    k = (sig1.s + c_1 * q) % ec.n
    q, _ = gen_keys(q)
    k, _ = gen_keys(k)
    return q, k


def crack_prv_key(
    msg1: String,
    sig1: Union[Sig, Octets],
    msg2: String,
    sig2: Union[Sig, Octets],
    Q: BIP340PubKey,
    hf: HashF = sha256,
) -> Tuple[int, int]:

    m_1 = reduce_to_hlen(msg1, hf)
    m_2 = reduce_to_hlen(msg2, hf)

    return _crack_prv_key(m_1, sig1, m_2, sig2, Q, hf)


def _assert_batch_as_valid(
    ms: Sequence[Octets],
    Qs: Sequence[BIP340PubKey],
    sigs: Sequence[Sig],
    hf: HashF = sha256,
) -> None:

    batch_size = len(Qs)
    if batch_size == 0:
        raise BTClibValueError("no signatures provided")

    if len(ms) != batch_size:
        err_msg = f"mismatch between number of pub_keys ({batch_size}) "
        err_msg += f"and number of messages ({len(ms)})"
        raise BTClibValueError(err_msg)
    if len(sigs) != batch_size:
        err_msg = f"mismatch between number of pub_keys ({batch_size}) "
        err_msg += f"and number of signatures ({len(sigs)})"
        raise BTClibValueError(err_msg)

    if batch_size == 1:
        _assert_as_valid(ms[0], Qs[0], sigs[0], hf)
        return None

    ec = sigs[0].ec
    if any(sig.ec != ec for sig in sigs):
        raise BTClibValueError("not the same curve for all signatures")
    t = 0
    scalars: List[int] = []
    points: List[JacPoint] = []
    for i, (m, Q, sig) in enumerate(zip(ms, Qs, sigs)):
        m = bytes_from_octets(m, hf().digest_size)

        KJ = sig.r, ec.y_even(sig.r), 1

        x_Q, y_Q = point_from_bip340pub_key(Q, ec)
        QJ = x_Q, y_Q, 1

        c = _challenge(m, x_Q, sig.r, ec, hf)

        # rand in [1, n-1]
        # deterministically generated using a CSPRNG seeded by a
        # cryptographic hash (e.g., SHA256) of all inputs of the
        # algorithm, or randomly generated independently for each
        # run of the batch verification algorithm
        rand = 1 if i == 0 else 1 + secrets.randbelow(ec.n - 1)
        scalars.append(rand)
        points.append(KJ)
        scalars.append(rand * c % ec.n)
        points.append(QJ)
        t += rand * sig.s

    TJ = _mult(t, ec.GJ, ec)
    RHSJ = _multi_mult(scalars, points, ec)

    # return T == RHS, checked in Jacobian coordinates
    RHSZ2 = RHSJ[2] * RHSJ[2]
    TZ2 = TJ[2] * TJ[2]
    if (TJ[0] * RHSZ2 % ec.p != RHSJ[0] * TZ2 % ec.p) or (
        TJ[1] * RHSZ2 * RHSJ[2] % ec.p != RHSJ[1] * TZ2 * TJ[2] % ec.p
    ):
        raise BTClibRuntimeError("signature verification failed")
    return None


def assert_batch_as_valid(
    ms: Sequence[String],
    Qs: Sequence[BIP340PubKey],
    sigs: Sequence[Sig],
    hf: HashF = sha256,
) -> None:

    ms = [reduce_to_hlen(m, hf) for m in ms]
    return _assert_batch_as_valid(ms, Qs, sigs, hf)


def _batch_verify(
    ms: Sequence[Octets],
    Qs: Sequence[BIP340PubKey],
    sigs: Sequence[Sig],
    hf: HashF = sha256,
) -> bool:

    # all kind of Exceptions are catched because
    # verify must always return a bool
    try:
        _assert_batch_as_valid(ms, Qs, sigs, hf)
    except Exception:  # pylint: disable=broad-except
        return False

    return True


def batch_verify(
    ms: Sequence[String],
    Qs: Sequence[BIP340PubKey],
    sigs: Sequence[Sig],
    hf: HashF = sha256,
) -> bool:
    """Batch verification of BIP340 signatures."""

    ms = [reduce_to_hlen(m, hf) for m in ms]
    return _batch_verify(ms, Qs, sigs, hf)



#!/usr/bin/env python3

# Copyright (C) 2020-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.tx_out` module."

import json
from os import path

import pytest

from btclib.exceptions import BTClibValueError
from btclib.tx import Tx
from btclib.tx_out import MAX_SATOSHI, TxOut


def test_tx_out() -> None:
    tx_out = TxOut()
    assert tx_out.value == 0
    assert tx_out.script_pub_key == b""
    assert tx_out.script_type == "unknown"
    assert tx_out.network == "mainnet"
    assert tx_out.address == ""
    assert tx_out.nValue == tx_out.value
    assert tx_out.scriptPubKey == tx_out.script_pub_key
    assert tx_out == TxOut.deserialize(tx_out.serialize())

    value = 3259343370
    script_ = bytes.fromhex(
        "0020ed8e9600561000f722bd26e850be7d80f24d174fabeff98baef967325e2b5a86"
    )
    tx_out = TxOut(value, script_)
    assert tx_out.value == value
    assert tx_out.script_pub_key == script_
    assert tx_out.script_type == "p2wsh"
    assert tx_out.network == "mainnet"
    addr = "bc1qak8fvqzkzqq0wg4aym59p0nasrey696040hlnzawl9nnyh3tt2rqzgmhmv"
    assert tx_out.address == addr
    assert tx_out.nValue == tx_out.value
    assert tx_out.scriptPubKey == tx_out.script_pub_key
    assert tx_out == TxOut.deserialize(tx_out.serialize())
    assert tx_out == TxOut.from_address(tx_out.value, tx_out.address)


def test_invalid_tx_out() -> None:
    script_pub_key = bytes.fromhex("6a0b68656c6c6f20776f726c64")
    tx_out = TxOut(-1, script_pub_key, check_validity=False)
    with pytest.raises(BTClibValueError, match="negative value: "):
        tx_out.assert_valid()

    tx_out = TxOut(MAX_SATOSHI + 1, script_pub_key, check_validity=False)
    with pytest.raises(BTClibValueError, match="value too high: "):
        tx_out.assert_valid()


def test_tx_out_from_address() -> None:
    address = "bc1qwqdg6squsna38e46795at95yu9atm8azzmyvckulcc7kytlcckxswvvzej"
    assert TxOut.from_address(0, address).address == address
    assert TxOut.from_address(0, address).network == "mainnet"
    address = "tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7"
    assert TxOut.from_address(0, address).address == address
    assert TxOut.from_address(0, address).network == "testnet"


def test_dataclasses_json_dict() -> None:
    fname = "d4f3c2c3c218be868c77ae31bedb497e2f908d6ee5bbbe91e4933e6da680c970.bin"
    filename = path.join(path.dirname(__file__), "test_data", fname)
    with open(filename, "rb") as binary_file_:
        temp = Tx.deserialize(binary_file_.read())

    tx_out_data = temp.vout[0]

    # dataclass
    assert isinstance(tx_out_data, TxOut)

    # Tx to/from dict
    tx_out_dict = tx_out_data.to_dict()
    assert isinstance(tx_out_dict, dict)
    assert tx_out_data == TxOut.from_dict(tx_out_dict)

    datadir = path.join(path.dirname(__file__), "generated_files")

    # Tx dict to/from dict file
    filename = path.join(datadir, "tx_out.json")
    with open(filename, "w") as file_:
        json.dump(tx_out_dict, file_, indent=4)
    with open(filename, "r") as file_:
        tx_dict2 = json.load(file_)
    assert isinstance(tx_dict2, dict)
    assert tx_out_dict == tx_dict2



#!/usr/bin/env python3

# Copyright (C) 2020-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

from dataclasses import InitVar, dataclass, field
from typing import Type, TypeVar

from dataclasses_json import DataClassJsonMixin, config

from . import var_bytes
from .alias import BinaryData
from .exceptions import BTClibValueError
from .utils import bytesio_from_binarydata
from .witness import Witness

_OutPoint = TypeVar("_OutPoint", bound="OutPoint")


@dataclass
class OutPoint(DataClassJsonMixin):
    tx_id: bytes = field(
        default=b"\x00" * 32,
        metadata=config(
            field_name="txid", encoder=lambda v: v.hex(), decoder=bytes.fromhex
        ),
    )
    vout: int = 0xFFFFFFFF
    # TODO add value and script_pub_key when tx fetcher will be available
    check_validity: InitVar[bool] = True

    @property
    def hash(self) -> int:
        "Return the hash int for compatibility with COutPoint."
        return int.from_bytes(self.tx_id, "big", signed=False)

    @property
    def n(self) -> int:
        "Return the n int for compatibility with COutPoint."
        return self.vout

    def __post_init__(self, check_validity: bool) -> None:
        if check_validity:
            self.assert_valid()

    def is_coinbase(self) -> bool:
        return self.tx_id == b"\x00" * 32 and self.vout == 0xFFFFFFFF

    def assert_valid(self) -> None:
        if len(self.tx_id) != 32:
            m = f"invalid OutPoint tx_id: {len(self.tx_id)}"
            m += " instead of 32 bytes"
            raise BTClibValueError(m)
        # must be a 4-bytes int
        if not 0 <= self.vout <= 0xFFFFFFFF:
            raise BTClibValueError(f"invalid vout: {self.vout}")
        # not a coinbase, not a regular OutPoint
        if (self.tx_id == b"\x00" * 32) ^ (self.vout == 0xFFFFFFFF):
            raise BTClibValueError("invalid OutPoint")

    def serialize(self, check_validity: bool = True) -> bytes:
        "Return the 36 bytes serialization of the OutPoint."

        if check_validity:
            self.assert_valid()

        out = self.tx_id[::-1]
        out += self.vout.to_bytes(4, byteorder="little", signed=False)
        return out

    @classmethod
    def deserialize(
        cls: Type[_OutPoint], data: BinaryData, check_validity: bool = True
    ) -> _OutPoint:
        "Return an OutPoint from the first 36 bytes of the provided data."

        data = bytesio_from_binarydata(data)
        tx_id = data.read(32)[::-1]
        vout = int.from_bytes(data.read(4), "little", signed=False)

        return cls(tx_id, vout, check_validity)


_TxIn = TypeVar("_TxIn", bound="TxIn")

_TX_IN_COMPARES_WITNESS = True


@dataclass
class TxIn(DataClassJsonMixin):
    prev_out: OutPoint = OutPoint()
    # TODO make it { "asm": "", "hex": "" }
    script_sig: bytes = field(
        default=b"",
        metadata=config(
            field_name="scriptSig", encoder=lambda v: v.hex(), decoder=bytes.fromhex
        ),
    )
    # If all TxIns have final (0xffffffff) sequence numbers
    # then Tx lock_time is irrelevant.
    #
    # Set to 0xFFFFFFFE to enables nLocktime (e.g. to discourage fee sniping)
    # and disables Replace-By-Fee (RBF).
    #
    # RBF txs typically have the sequence of each input set to 0xFFFFFFFD.
    #
    # Because sequence locks require that the sequence field be set
    # lower than 0xFFFFFFFD to be meaningful,
    # all sequence locked transactions are opting into RBF.
    sequence: int = 0
    script_witness: Witness = field(
        default=Witness(),
        init=True,  # must be True, probably a bug of dataclasses_json
        repr=True,
        compare=_TX_IN_COMPARES_WITNESS,
        metadata=config(field_name="txinwitness"),
    )
    check_validity: InitVar[bool] = True

    @property
    def outpoint(self) -> OutPoint:
        "Return the outpoint OutPoint for compatibility with CTxIn."
        return self.prev_out

    @property
    def scriptSig(self) -> bytes:  # pylint: disable=invalid-name
        "Return the scriptSig bytes for compatibility with CTxIn."
        return self.script_sig

    @property
    def nSequence(self) -> int:  # pylint: disable=invalid-name
        "Return the nSequence int for compatibility with CTxIn."
        return self.sequence

    def __post_init__(self, check_validity: bool) -> None:
        if check_validity:
            self.assert_valid()

    def is_segwit(self) -> bool:
        # self.prev_out has no segwit information
        return self.script_witness.stack != []

    def is_coinbase(self) -> bool:
        return self.prev_out.is_coinbase()

    def assert_valid(self) -> None:
        self.prev_out.assert_valid()

        # TODO check script_sig

        # must be a 4-bytes int
        if not 0 <= self.sequence <= 0xFFFFFFFF:
            raise BTClibValueError(f"invalid sequence: {self.sequence}")

        if self.script_witness:
            self.script_witness.assert_valid()

    def serialize(self, check_validity: bool = True) -> bytes:

        if check_validity:
            self.assert_valid()

        out = self.prev_out.serialize()
        out += var_bytes.serialize(self.script_sig)
        out += self.sequence.to_bytes(4, byteorder="little", signed=False)
        return out

    @classmethod
    def deserialize(
        cls: Type[_TxIn], data: BinaryData, check_validity: bool = True
    ) -> _TxIn:

        s = bytesio_from_binarydata(data)
        prev_out = OutPoint.deserialize(s)
        script_sig = var_bytes.deserialize(s)
        sequence = int.from_bytes(s.read(4), byteorder="little", signed=False)

        return cls(prev_out, script_sig, sequence, Witness(), check_validity)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Aliases

mypy aliases, documenting also coding imput conventions.
"""

from io import BytesIO
from typing import Any, Callable, Tuple, Union

# Octets are a sequence of eight-bit bytes or a hex-string (not text string)
#
# hex-strings are strings that can be converted to bytes using bytes.fromhex,
# e.g.:
# "deadbeef"
# "dead beef"
# "04 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf f7d8a473e7e2e6d317b87bafe8bde97e3cf8f065dec022b51d11fcdd0d348ac4"
# "02 cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
# "02cc71eb30d653c0c3163990c47b976f3fb3f37cccdcbedb169a1dfef58bbfbfaf"
#
# use btclib.utils.bytes_from_octets to convert Octets to bytes
#
# Octets are used for serialized script, h160 (20 bytes), h256 (32 bytes),
# BIP32 version (4 bytes), sig_hash_type (1 byte),
# dsa.Sig (DER serialization of ECDSA signature),
# ssa.Sig (BIP340 serialization of Schnorr signature)
# etc.
Octets = Union[bytes, str]

# bytes or text string (not hex-string)
#
# this is for string that can be
# converted to bytes using encode()
# e.g. a message to be signed
#    if isinstance(msg, str):
#        msg = msg.encode()
#
# or 'ascii' strings like addresses (base58 or bech32),
# WIFs, or BIP32 keys:
# "37k7toV1Nv4DfmQbmZ8KuZDQCYK9x5KpzP"
# "KyLk7s6Z1FtgYEVp3bPckPVnXvLUWNCcVL6wNt3gaT96EmzTKZwP"
# "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
# "bc1qg9stkxrszkdqsuj92lm4c7akvk36zvhqw7p6ck"
#
# also bms.Sig (Bitcoin message compact signature serialization),
#
# In almost all cases (but messages to be signed)
# leading/trailing blanks should always be stripped
#     if isinstance(b58addr, str):
#         b58addr = b58addr.strip()
#
# In those cases often there is no need to encode() to bytes
# as b58decode/b32decode/etc. will take care of that
String = Union[bytes, str]

# binary data, usually to be cosumed as byte stream,
# but possibily provided as Octets too
BinaryData = Union[BytesIO, Octets]

# hex-string or bytes representation of an int
# Integer = Union[Octets, int]
Integer = Union[bytes, str, int]

# Hash digest constructor: it may be any name suitable to hashlib.new()
HashF = Callable[[], Any]
# HashF = Callable[[Any], Any]

# Elliptic curve point in affine coordinates.
# Warning: to make Point a NamedTuple would slow down the code
Point = Tuple[int, int]

# Note that the infinity point in affine coordinates is INF = (int, 0)
# (no affine point has y=0 coordinate in a group of prime order n).
# It can be checked with 'INF[1] == 0'
# The x-coordinate is arbitrary: 7 is preferred
# because it is not a field element in secp256k1
INF = 7, 0

# Elliptic curve point in Jacobian coordinates.
JacPoint = Tuple[int, int, int]

# Infinity point in Jacobian coordinates is INF = (int, int, 0).
# It can be checked with 'INF[2] == 0'
# The default x and y coordinates are arbitrary:
# 7, 0 are used because those are what one would obtain
# from the generic affine to Jacobian transformation
# of the INF Point
# QJ = Q[0], Q[1], 1 if Q[1] else 0
INFJ = 7, 0, 0

# the main internal representation of entropy is binary 0/1 string
BinStr = str
# but int or bytes are fine too
Entropy = Union[BinStr, int, bytes]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic Curve Digital Signature Algorithm (ECDSA).

   Implementation according to SEC 1 v.2:

   http://www.secg.org/sec1-v2.pdf

   specialized with bitcoin canonical 'low-s' encoding.
"""

import secrets
from hashlib import sha256
from typing import List, Optional, Tuple, Union

from .alias import HashF, JacPoint, Octets, Point, String
from .curve import Curve, secp256k1
from .curve_group import _double_mult, _mult
from .der import Sig
from .exceptions import BTClibRuntimeError, BTClibValueError
from .hashes import reduce_to_hlen
from .number_theory import mod_inv
from .rfc6979 import __rfc6979
from .to_prv_key import PrvKey, int_from_prv_key
from .to_pub_key import Key, point_from_key
from .utils import bytes_from_octets, int_from_bits


def gen_keys(prv_key: PrvKey = None, ec: Curve = secp256k1) -> Tuple[int, Point]:
    "Return a private/public (int, Point) key-pair."

    if prv_key is None:
        # q in the range [1, ec.n-1]
        q = 1 + secrets.randbelow(ec.n - 1)
    else:
        q = int_from_prv_key(prv_key, ec)

    QJ = _mult(q, ec.GJ, ec)
    Q = ec._aff_from_jac(QJ)
    # q.to_bytes(ec.nsize, byteorder="big", signed=False)
    # bytes_from_point(Q, ec, compressed)
    return q, Q


def _challenge(m: Octets, ec: Curve = secp256k1, hf: HashF = sha256) -> int:

    # The message m: a hlen array
    hlen = hf().digest_size
    m = bytes_from_octets(m, hlen)

    # leftmost ec.nlen bits %= ec.n
    c = int_from_bits(m, ec.nlen) % ec.n  # 5
    return c


def challenge(msg: String, ec: Curve = secp256k1, hf: HashF = sha256) -> int:

    m = reduce_to_hlen(msg, hf)
    return _challenge(m, ec, hf)


def __sign(c: int, q: int, k: int, low_s: bool, ec: Curve) -> Sig:
    # Private function for testing purposes: it allows to explore all
    # possible value of the challenge c (for low-cardinality curves).
    # It assume that c is in [0, n-1], while q and k are in [1, n-1]

    # Steps numbering follows SEC 1 v.2 section 4.1.3

    KJ = _mult(k, ec.GJ, ec)  # 1

    # affine x_K-coordinate of K (field element)
    x_K = (KJ[0] * mod_inv(KJ[2] * KJ[2], ec.p)) % ec.p
    # mod n makes it a scalar
    r = x_K % ec.n  # 2, 3
    if r == 0:  # r≠0 required as it multiplies the public key
        raise BTClibRuntimeError("failed to sign: r = 0")

    s = mod_inv(k, ec.n) * (c + r * q) % ec.n  # 6
    if s == 0:  # s≠0 required as verify will need the inverse of s
        raise BTClibRuntimeError("failed to sign: s = 0")

    # bitcoin canonical 'low-s' encoding for ECDSA signatures
    # it removes signature malleability as cause of transaction malleability
    # see https://github.com/bitcoin/bitcoin/pull/6769
    if low_s and s > ec.n / 2:
        s = ec.n - s  # s = - s % ec.n

    return Sig(r, s, ec)


def _sign(
    m: Octets,
    prv_key: PrvKey,
    k: Optional[PrvKey] = None,
    low_s: bool = True,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> Sig:
    """Sign a hlen bytes message according to ECDSA signature algorithm.

    If the deterministic nonce is not provided,
    the RFC6979 specification is used.
    """

    # the message m: a hlen array
    hlen = hf().digest_size
    m = bytes_from_octets(m, hlen)

    # the secret key q: an integer in the range 1..n-1.
    # SEC 1 v.2 section 3.2.1
    q = int_from_prv_key(prv_key, ec)

    # the challenge
    c = _challenge(m, ec, hf)  # 4, 5

    # the nonce k: an integer in the range 1..n-1.
    if k is None:
        k = __rfc6979(c, q, ec, hf)  # 1
    else:
        k = int_from_prv_key(k, ec)

    # second part delegated to helper function
    return __sign(c, q, k, low_s, ec)


def sign(
    msg: String,
    prv_key: PrvKey,
    low_s: bool = True,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> Sig:
    """ECDSA signature with canonical low-s preference.

    Implemented according to SEC 1 v.2
    The message msg is first processed by hf, yielding the value

        m = hf(msg),

    a sequence of bits of length *hlen*.

    Normally, hf is chosen such that its output length *hlen* is
    roughly equal to *nlen*, the bit-length of the group order *n*,
    since the overall security of the signature scheme will depend on
    the smallest of *hlen* and *nlen*; however, the ECDSA standard
    supports all combinations of *hlen* and *nlen*.

    RFC6979 is used for deterministic nonce.

    See https://tools.ietf.org/html/rfc6979#section-3.2
    """

    m = reduce_to_hlen(msg, hf)
    return _sign(m, prv_key, None, low_s, ec, hf)


def __assert_as_valid(c: int, QJ: JacPoint, r: int, s: int, ec: Curve) -> None:
    # Private function for test/dev purposes

    w = mod_inv(s, ec.n)
    u = c * w % ec.n
    v = r * w % ec.n  # 4
    # Let K = u*G + v*Q.
    KJ = _double_mult(v, QJ, u, ec.GJ, ec)  # 5

    # Fail if infinite(K).
    # edge case that cannot be reproduced in the test suite
    if KJ[2] == 0:  # 5
        err_msg = "invalid (INF) key"  # pragma: no cover
        raise BTClibRuntimeError(err_msg)  # pragma: no cover

    # affine x_K-coordinate of K
    x_K = (KJ[0] * mod_inv(KJ[2] * KJ[2], ec.p)) % ec.p
    # Fail if r ≠ x_K %n.
    if r != x_K % ec.n:  # 6, 7, 8
        raise BTClibRuntimeError("signature verification failed")


def _assert_as_valid(
    m: Octets, key: Key, sig: Union[Sig, Octets], hf: HashF = sha256
) -> None:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    if not isinstance(sig, Sig):
        sig = Sig.deserialize(sig)
    else:
        sig.assert_valid()  # 1

    # The message m: a hlen array
    m = bytes_from_octets(m, hf().digest_size)
    c = _challenge(m, sig.ec, hf)  # 2, 3

    Q = point_from_key(key, sig.ec)
    QJ = Q[0], Q[1], 1

    # second part delegated to helper function
    __assert_as_valid(c, QJ, sig.r, sig.s, sig.ec)


def assert_as_valid(
    msg: String, key: Key, sig: Union[Sig, Octets], hf: HashF = sha256
) -> None:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    m = reduce_to_hlen(msg, hf)
    _assert_as_valid(m, key, sig, hf)


def _verify(m: Octets, key: Key, sig: Union[Sig, Octets], hf: HashF = sha256) -> bool:
    """ECDSA signature verification (SEC 1 v.2 section 4.1.4)."""

    # all kind of Exceptions are catched because
    # verify must always return a bool
    try:
        _assert_as_valid(m, key, sig, hf)
    except Exception:  # pylint: disable=broad-except
        return False
    else:
        return True


def verify(msg: String, key: Key, sig: Union[Sig, Octets], hf: HashF = sha256) -> bool:
    """ECDSA signature verification (SEC 1 v.2 section 4.1.4)."""

    m = reduce_to_hlen(msg, hf)
    return _verify(m, key, sig, hf)


def recover_pub_keys(
    msg: String, sig: Union[Sig, Octets], hf: HashF = sha256
) -> List[Point]:
    """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).

    See also:
    https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
    """

    m = reduce_to_hlen(msg, hf)
    return _recover_pub_keys(m, sig, hf)


def _recover_pub_keys(
    m: Octets, sig: Union[Sig, Octets], hf: HashF = sha256
) -> List[Point]:
    """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).

    See also:
    https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
    """

    if not isinstance(sig, Sig):
        sig = Sig.deserialize(sig)
    else:
        sig.assert_valid()  # 1

    # The message m: a hlen array
    hlen = hf().digest_size
    m = bytes_from_octets(m, hlen)

    c = _challenge(m, sig.ec, hf)  # 1.5

    QJs = __recover_pub_keys(c, sig.r, sig.s, sig.ec)
    return [sig.ec._aff_from_jac(QJ) for QJ in QJs]


# TODO: use __recover_pub_key to avoid code duplication
def __recover_pub_keys(c: int, r: int, s: int, ec: Curve) -> List[JacPoint]:
    # Private function provided for testing purposes only.

    # precomputations
    r_1 = mod_inv(r, ec.n)
    r1s = r_1 * s % ec.n
    r1e = -r_1 * c % ec.n
    keys: List[JacPoint] = []
    # r = K[0] % ec.n
    # if ec.n < K[0] < ec.p (likely when cofactor ec.cofactor > 1)
    # then both x_K=r and x_K=r+ec.n must be tested
    for j in range(ec.cofactor + 1):  # 1
        # affine x_K-coordinate of K (field element)
        x_K = (r + j * ec.n) % ec.p  # 1.1
        # two possible y_K-coordinates, i.e. two possible keys for each cycle
        try:
            # even root first for bitcoin message signing compatibility
            yodd = ec.y_even(x_K)
            KJ = x_K, yodd, 1  # 1.2, 1.3, and 1.4
            # 1.5 has been performed in the recover_pub_keys calling function
            QJ = _double_mult(r1s, KJ, r1e, ec.GJ, ec)  # 1.6.1
            try:
                __assert_as_valid(c, QJ, r, s, ec)  # 1.6.2
            except (BTClibValueError, BTClibRuntimeError):
                pass
            else:
                keys.append(QJ)  # 1.6.2
            KJ = x_K, ec.p - yodd, 1  # 1.6.3
            QJ = _double_mult(r1s, KJ, r1e, ec.GJ, ec)
            try:
                __assert_as_valid(c, QJ, r, s, ec)  # 1.6.2
            except (BTClibValueError, BTClibRuntimeError):
                pass
            else:
                keys.append(QJ)  # 1.6.2
        except (BTClibValueError, BTClibRuntimeError):  # K is not a curve point
            pass
    return keys


def __recover_pub_key(key_id: int, c: int, r: int, s: int, ec: Curve) -> JacPoint:
    # Private function provided for testing purposes only.

    # precomputations
    r_1 = mod_inv(r, ec.n)
    r1s = r_1 * s % ec.n
    r1e = -r_1 * c % ec.n
    # r = K[0] % ec.n
    # if ec.n < K[0] < ec.p (likely when cofactor ec.cofactor > 1)
    # then both x_K=r and x_K=r+ec.n must be tested
    j = key_id & 0b110  # allow for key_id in [0, 7]
    x_K = (r + j * ec.n) % ec.p  # 1.1

    # even root first for Bitcoin Core compatibility
    i = key_id & 0b01
    y_even = ec.y_even(x_K)
    y_K = ec.p - y_even if i else y_even
    KJ = x_K, y_K, 1  # 1.2, 1.3, and 1.4
    # 1.5 has been performed in the recover_pub_keys calling function
    QJ = _double_mult(r1s, KJ, r1e, ec.GJ, ec)  # 1.6.1
    __assert_as_valid(c, QJ, r, s, ec)  # 1.6.2
    return QJ


def _crack_prv_key(
    m_1: Octets,
    sig1: Union[Sig, Octets],
    m_2: Octets,
    sig2: Union[Sig, Octets],
    hf: HashF = sha256,
) -> Tuple[int, int]:

    if not isinstance(sig1, Sig):
        sig1 = Sig.deserialize(sig1)
    else:
        sig1.assert_valid()  # 1

    if not isinstance(sig2, Sig):
        sig2 = Sig.deserialize(sig2)
    else:
        sig2.assert_valid()  # 1

    ec = sig2.ec
    if sig1.ec != ec:
        raise BTClibValueError("not the same curve in signatures")
    if sig1.r != sig2.r:
        raise BTClibValueError("not the same r in signatures")
    if sig1.s == sig2.s:
        raise BTClibValueError("identical signatures")

    hlen = hf().digest_size
    m_1 = bytes_from_octets(m_1, hlen)
    m_2 = bytes_from_octets(m_2, hlen)

    c_1 = _challenge(m_1, ec, hf)
    c_2 = _challenge(m_2, ec, hf)

    k = (c_1 - c_2) * mod_inv(sig1.s - sig2.s, ec.n) % ec.n
    q = (sig2.s * k - c_2) * mod_inv(sig1.r, ec.n) % ec.n
    return q, k


def crack_prv_key(
    msg1: String,
    sig1: Union[Sig, Octets],
    msg2: String,
    sig2: Union[Sig, Octets],
    hf: HashF = sha256,
) -> Tuple[int, int]:

    m_1 = reduce_to_hlen(msg1, hf)
    m_2 = reduce_to_hlen(msg2, hf)

    return _crack_prv_key(m_1, sig1, m_2, sig2, hf)



#!/usr/bin/env python3

# Copyright (C) 2020-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

from dataclasses import InitVar, dataclass, field
from typing import List, Type, TypeVar

from dataclasses_json import DataClassJsonMixin, config

from . import var_bytes, var_int
from .alias import BinaryData
from .exceptions import BTClibTypeError
from .utils import bytesio_from_binarydata

_TxInWitness = TypeVar("_TxInWitness", bound="Witness")


@dataclass
class Witness(DataClassJsonMixin):
    stack: List[bytes] = field(
        default_factory=list,
        metadata=config(
            encoder=lambda val: [v.hex() for v in val],
            decoder=lambda val: [bytes.fromhex(v) for v in val],
        ),
    )
    check_validity: InitVar[bool] = True

    def __post_init__(self, check_validity: bool) -> None:
        if check_validity:
            self.assert_valid()

    def __len__(self):
        return len(self.stack)

    def assert_valid(self) -> None:
        if not isinstance(self.stack, list):
            raise BTClibTypeError("invalid witness")
        if not all(isinstance(v, bytes) for v in self.stack):
            raise BTClibTypeError("invalid witness")

    def serialize(self, check_validity: bool = True) -> bytes:
        "Return the serialization of the Witness."

        if check_validity:
            self.assert_valid()

        out = var_int.serialize(len(self.stack))
        return out + b"".join([var_bytes.serialize(w) for w in self.stack])

    @classmethod
    def deserialize(
        cls: Type[_TxInWitness], data: BinaryData, check_validity: bool = True
    ) -> _TxInWitness:
        "Return a Witness by parsing binary data."

        data = bytesio_from_binarydata(data)
        n = var_int.deserialize(data)
        stack = [var_bytes.deserialize(data) for _ in range(n)]
        return cls(stack, check_validity)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

""" Hash based helper functions.

"""

import hashlib
from typing import Optional, Tuple

from .alias import HashF, String
from .to_pub_key import Key, pub_keyinfo_from_key
from .utils import hash160

H160_Net = Tuple[bytes, str]


def hash160_from_key(
    key: Key, network: Optional[str] = None, compressed: Optional[bool] = None
) -> H160_Net:
    """Return (public key HASH160, nettwork) from a private/public key.

    HASH160 is RIPEMD160(SHA256).
    """
    pub_key, network = pub_keyinfo_from_key(key, network, compressed)
    return hash160(pub_key), network


def fingerprint(key: Key, network: Optional[str] = None) -> bytes:
    """Return the public key fingerprint from a private/public key.

    The fingerprint is the last four bytes
    of the compressed public key HASH160.
    """

    pub_key, _ = pub_keyinfo_from_key(key, network, compressed=True)
    return hash160(pub_key)[:4]


def reduce_to_hlen(msg: String, hf: HashF = hashlib.sha256) -> bytes:

    if isinstance(msg, str):
        # do not strip spaces
        msg = msg.encode()

    # Step 4 of SEC 1 v.2 section 4.1.3
    h = hf()
    h.update(msg)
    return h.digest()


def tagged_hash(tag: str, msg: String, hf: HashF = hashlib.sha256) -> bytes:

    if isinstance(msg, str):
        # do not strip spaces
        msg = msg.encode()

    t = tag.encode()
    h1 = hf()
    h1.update(t)
    tag_hash = h1.digest()

    h2 = hf()
    h2.update(tag_hash + tag_hash)

    # it could be sped up by storing the above midstate
    h2.update(msg)
    return h2.digest()



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"Tests for the `btclib.script` module."

from typing import List

import pytest

from btclib import script
from btclib.exceptions import BTClibTypeError, BTClibValueError


def test_operators() -> None:
    for i in script.OP_CODE_NAMES:
        b = script.OP_CODES[script.OP_CODE_NAMES[i]]
        assert i == b[0]
    for name in script.OP_CODES:
        # skip duplicated
        if name in ("OP_FALSE", "OP_TRUE", "OP_NOP2", "OP_NOP3"):
            continue
        i = script.OP_CODES[name][0]
        assert name == script.OP_CODE_NAMES[i]
    for i in range(76, 186):
        # skip disabled 'splice' opcodes
        if i in (126, 127, 128, 129):
            continue
        # skip disabled 'bitwise logic' opcodes
        if i in (131, 132, 133, 134):
            continue
        # skip disabled 'splice' opcodes
        if i in (141, 142, 149, 150, 151, 152, 152, 153):
            continue
        # skip 'reserved' opcodes
        if i in (80, 98, 101, 102, 137, 138):
            continue
        assert i in script.OP_CODE_NAMES.keys()


def test_simple() -> None:
    script_list: List[List[script.ScriptToken]] = [
        [2, 3, "OP_ADD", 5, "OP_EQUAL"],
        ["1ADD", "OP_1ADD", "1ADE", "OP_EQUAL"],
        [hex(26)[2:].upper(), -1, "OP_ADD", hex(26)[2:].upper(), "OP_EQUAL"],
        [
            hex(0xFFFFFFFF)[2:].upper(),
            -1,
            "OP_ADD",
            hex(0xFFFFFFFF)[2:].upper(),
            "OP_EQUAL",
        ],
        ["1F" * 250, "OP_DROP"],
        ["1F" * 520, "OP_DROP"],
    ]
    for script_pub_key in script_list:
        assert script_pub_key == script.deserialize(script.serialize(script_pub_key))
        assert script_pub_key == script.deserialize(
            script.serialize(script_pub_key).hex()
        )


def test_exceptions() -> None:

    script_pub_key: List[script.ScriptToken] = [2, 3, "OP_ADD", 5, "OP_RETURN_244"]
    err_msg = "invalid string token: OP_RETURN_244"
    with pytest.raises(BTClibValueError, match=err_msg):
        script.serialize(script_pub_key)

    err_msg = "Unmanaged <class 'function'> token type"
    with pytest.raises(BTClibTypeError, match=err_msg):
        script.serialize([2, 3, "OP_ADD", 5, script.serialize])  # type: ignore

    err_msg = "Too many bytes for OP_PUSHDATA: "
    with pytest.raises(BTClibValueError, match=err_msg):
        script_pub_key = ["1f" * 521, "OP_DROP"]
        script.serialize(script_pub_key)

    # A script_pub_key with OP_PUSHDATA4 can be decoded
    script_bytes = "4e09020000" + "00" * 521 + "75"  # ['00'*521, 'OP_DROP']
    script_pub_key_ = script.deserialize(script_bytes)
    # but it cannot be encoded
    err_msg = "Too many bytes for OP_PUSHDATA: "
    with pytest.raises(BTClibValueError, match=err_msg):
        script.serialize(script_pub_key_)


def test_nulldata() -> None:

    scripts: List[List[script.ScriptToken]] = [
        ["OP_RETURN", "11" * 79],
        ["OP_RETURN", "00" * 79],
    ]
    for script_pub_key in scripts:
        assert script_pub_key == script.deserialize(script.serialize(script_pub_key))
        assert script_pub_key == script.deserialize(
            script.serialize(script_pub_key).hex()
        )


def test_op_int() -> None:
    i = 0b01111111
    assert len(script._op_int(i)) == 2
    i = 0b11111111
    assert len(script._op_int(i)) == 3
    i = 0b0111111111111111
    assert len(script._op_int(i)) == 3
    i = 0b1111111111111111
    assert len(script._op_int(i)) == 4


def test_op_pushdata() -> None:
    length = 75
    b = "00" * length
    assert len(script._op_pushdata(b)) == length + 1
    b = "00" * (length + 1)
    assert len(script._op_pushdata(b)) == (length + 1) + 2

    length = 255
    b = "00" * length
    assert len(script._op_pushdata(b)) == length + 2
    b = "00" * (length + 1)
    assert len(script._op_pushdata(b)) == (length + 1) + 3


def test_encoding():
    script_bytes = b"jKBIP141 \\o/ Hello SegWit :-) keep it strong! LLAP Bitcoin twitter.com/khs9ne"
    assert script.serialize(script.deserialize(script_bytes)) == script_bytes
